#include "PCGenMemo.h"

/* FUNCTION: makeMemo
 * PURPOSE: This function is used to create memos automatically. Given a memo number, *
 * a memo size and a file handle to random data, this function will generate memos of *
 * the following format: "MEMOXXX YYY...YYY" where XXX is the three digit field       *
 * determined by the memo number, and YYY...YYY is of the length specified in the size*
 * parameter. The actuall characters of YYY...YYY can be manipulated by a #define     *
 * statement. If there is a #define JOKESFILE, (and presumably you have passed the    *
 * FILE handle of that jokes file to this function) then the memos will not conform to*
 * the size parameter, instead, they will contain varaible length strings, depending  *
 * on what is in the FILE handle that is passed to the function. The return value is  *
 * a string containing the created memo.                                              *
 * DATE: 5/29/02                                                                      *
 * CONSTANTS: MEMOSIZE                                                                *
 * MODULES: PCGenMemo                                                                 *
 * SUGGESTIONS: The 2nd static char buffer[MEMOSIZE] declaration can be eliminated by *
 * moving them both out of the #if statements. So can other statements.               */
char * makeMemo(int memonumber, int size, FILE *randomdata )
{
#ifndef JOKESFILE							//if we do not want variable length memos

	static char buffer[MEMOSIZE]= "";
	unsigned short y, insert;
	
	sprintf(buffer, "MEMO#%.4d    ", memonumber );		//create the MEMOXXX part
	
	insert = strlen(buffer);
	for( y = 0; y < size; y++ )							//copy characters from the file for the rest
		fscanf( randomdata, "%c", &buffer[insert++] ); 

	buffer[insert] = '\0';

#endif

#ifdef JOKESFILE										//if we DO want variable length memos

	static char buffer[MEMOSIZE] = "";
	unsigned short insert;

	sprintf(buffer, "MEMO#%.3d    ", memonumber );		//create the MEMOXXX part
	insert = strlen(buffer);
	
	fgets( &buffer[insert], MEMOSIZE-10, randomdata );	//copy a string from the file, up until a '\0' or MEMOSIZE-10

	buffer[strlen(buffer)-1] = '\0';

#endif

	return( buffer );
}


/* FUNCTION: copyMemoEntry                                                                 *
 * PURPOSE: Given the pointers to a source and destination memos for a copy, this function *
 * deletes the destination memo (deallocates its resources) and then copies the source's   *
 * data onto it. It then deletes the source memo and deallocates its resources.            *
 * DATE: 6/3/02                                                                            *
 * CONSTANTS: none                                                                         *
 * MODULES: PCGenMemo                                                                      *
 * SUGGESTIONS: This function should really be called moveMemoEntry since it deletes the   *
 * source location.                                                                        */
void copyMemoEntry(memo *source, memo *dest)
{
	if( dest ->data != NULL )							//first, deallocate anything that may be in the destination already
		delete( dest ->data );

	dest ->data = new char[ strlen( source->data) + 1 ];	//then, reallocate space for the source's data
	strcpy( dest ->data, source->data );					//copy the source to the new destination
	dest ->hash = source->hash;								

	delete( source ->data );							//then delete the source
	source ->data = NULL;
	source ->hash = 0;
}


/* FUNCTION: calcHashValue                                               *
 * PURPOSE: Given a pointer to a character string, this function will    *
 * return the 16bit hash value for it. The hash function implemented     *
 * is simply the xor(?) of each even byte for the least significant byte *
 * of the hash value, and the xor(?) of each odd byte for the most       *
 * significant byte of the hash value.                                   *
 * DATE: 5/29/02                                                         *
 * CONSTANTS: DIFFERENCES                                                *
 *            256 - MAX 8bit value                                       *
 *            32749 - Prime field                                        *
 * MODULES: PCGenMemo                                                    */
unsigned short calcHashValue( const char *data)
{
	char hash[2] = {0,0};			//these two 8bit chars form the space for our 16bit hash value
	unsigned short *hashval;
	int temp;

	temp = strlen(data);
	if( temp % 2 )					//we must check the length so that we know whether the string 
	{								//ends on an odd or even character
		for( int x = 0; x < temp-1; x += 2 )
		{
			hash[0] ^= data[x];
			hash[1] ^= data[x+1];
		}
		hash[0] ^= data[x];
	}
	else
		for( int x = 0; x < temp; x += 2 )
		{
			hash[0] ^= data[x];
			hash[1] ^= data[x+1];
		}

	hashval = (unsigned short*)hash;
	*hashval = *hashval % 32749;
	
	//avoid collision between hash value and sample points (collision is when sample-hash = 0 or sample-hash = prime field)
	//the sample points go as {0, -1, 1, -2, 2... -N/2} for N number of differences so adding half the value of N (or 
	//subtracting it) should avoid collisions
	if( *hashval < DIFFERENCES/2 + 1 )
		*hashval += DIFFERENCES/2 + 1;
	else if( *hashval > 32749 - (DIFFERENCES/2 + 1) )
		*hashval -= DIFFERENCES/2 + 1;
	
	return( *hashval );
}


/* FUNCTION: createMemo                                                   *
 * PURPOSE: Given a string and a hash value and a pointer to the memo     *
 * data array, this function will add a memo at the end of the memo       *
 * array. It will also increment the value of nummemos that it was given. *
 * DATE: 6/21/02                                                          *
 * CONSTANTS: none                                                        *
 * MODULES: PCGenMemo                                                     */
void createMemo( const char *memotext, int hashvalue, int *nummemos, memo *memodata )
{
	memodata[*nummemos].data = new char[strlen(memotext)+1];	//copy the memo data to the PC array
	strcpy( memodata[*nummemos].data, memotext );
	memodata[*nummemos].hash = hashvalue;						//copy the memos hash value to the PC array
	(*nummemos)++;												//make sure to keep track of number of memos on PC
}


/* FUNCTION: freeData                                                 *
 * PURPOSE: Given a pointer to an array of type "memo", this function *
 * will go through the first NUMMEMOS indexes and free the memory     *
 * allocated to each memo string.                                     *
 * DATE: 6/3/02                                                       *
 * CONSTANTS: NUMMEMOS                                                *
 * MODULES: PCGenMemo                                                 */
void freeData(memo *memodata)
{
	for( int x = 0; x < NUMMEMOS; x++ )
		if( memodata[x].data != NULL )	
			delete( memodata[x].data );
}


/* FUNCTION: getMemoText                                               * 
 * PURPOSE: Given a pointer to the memo array and an index, this       * 
 * function will return a pointer to the memo text at the given index. *
 * DATE: 6/21/02                                                       *
 * CONSTANTS: none                                                     *
 * MODULES: PCGenMemo                                                  */
char *getMemoText( memo *memodata, int index )
{
	return( memodata[index].data );
}

/* FUNCTION: initializeNewData                                        *
 * PURPOSE: Given a pointer to an array of type "memo", this function *
 * will go through the first NUMMEMOS indexes and set the memo string *
 * pointer to null and the hashvalue to zero.                         *
 * DATE: 6/3/02                                                       *
 * CONSTANTS: NUMMEMOS                                                *
 * MODULES: PCGenMemo                                                 */
void initializeNewData( memo *memodata)
{
	for( int x = 0; x < NUMMEMOS; x++ )
	{
		memodata[x].data = NULL;
		memodata[x].hash = 0;
	}
}
