#include "\Common\Constants.h"

#ifdef HASH1					//only use this file if the array style hash is desired from constants.h

#include "\Palm\PalmHashDb\PalmHashDb.h"


/* FUNCTION: blockInsertHashIndexes                                                                              *
 * PURPOSE: Given the number of hashes passed to the function, an array of those hash structures and a reference *
 * to the pre-opened hash database, this function will insert all the hash values in one call. This is supposed  *
 * to save time over calling the insertHashIndex function over and over with a large list of hashes. As it is    *
 * implemented now, the primary savings over the insertHashIndex function is just in the allocation of the new   *
 * memory, locking of the hash record and deallocation/delocking of the same items.                              *
 * DATE: 5/28/02                                                                                                 *
 * CONSTANTS: NUMMEMOS                                                                                           *
 *            0 - the index of the record in the hash database that has all the hash values                      *
 * MODULES: PalmHashDb                                                                                           *
 * SUGGESTIONS: This function could be greatly improved if it could assume that the list of hashvalues it is     *
 * given had some order to them. Then it would only need to construct the buffer array once, and copy it once.   */
void blockInsertHashIndexes( UInt16 numhashes, HashRecordType hashvalues[], DmOpenRef hashDb )
{
	MemHandle record, newbuffer;
	HashRecordType *hashrecords, *buffer, hashdata;
	UInt16 index;
	Int16 x;
	
	//the insert will require an actual temp copy of the array, so allocate some heap for it, and lock it.
	newbuffer = MemHandleNew( NUMMEMOS * sizeof(HashRecordType) );
	buffer = MemHandleLock( newbuffer );
	
	record = DmGetRecord( hashDb, 0 );					//get the hash record array and lock it.
	hashrecords = MemHandleLock( record );
	
	for( x = numhashes-1; x >= 0; x-- )
	{
		hashdata = hashvalues[x];
		index = calcNewIndex( hashdata.hashval, hashDb );	//find the index in the array to insert the new hash data

		//first, move all the array before the new index to be inserted to the temporary buffer
		MemMove( buffer, hashrecords, index * sizeof(HashRecordType) );
		
		//second, move all the array after the new index to be inserted into the temp buffer ONE INDEX AHEAD. 
		MemMove( &(buffer[index+1]), &hashrecords[index], (hashrecords[0].docindex - index + 1) * sizeof(HashRecordType));
		
		buffer[index].docindex = hashdata.docindex;				//add the new index to its place in the temp buffer
		buffer[index].hashval = hashdata.hashval;
		buffer[0].docindex = hashrecords[0].docindex + 1;		//update the zeroth element in the temp buffer
		buffer[0].hashval = 0;
		DmWrite( hashrecords, 0, buffer, (buffer[0].docindex + 1) * sizeof(HashRecordType) );	//copy the temp buffer to the 
	}																						//array

	MemHandleUnlock( record );							//unlock and release the hash record
	DmReleaseRecord( hashDb, 0, true );

	MemHandleUnlock( newbuffer );						//free the dynamically allocated memory
	MemHandleFree( newbuffer );
}


/* FUNCTION: calcHashValue                                               *
 * PURPOSE: Given a pointer to a character string, this function will    *
 * return the 16bit hash value for it. The hash function implemented     *
 * is simply the xor(?) of each even byte for the least significant byte *
 * of the hash value, and the xor(?) of each odd byte for the most       *
 * significant byte of the hash value.                                   *
 * DATE: 8/16/02                                                         *
 * CONSTANTS: DIFFERENCES                                                *
 *            256 - MAX 8bit value                                       *
 *            32749 - Prime field                                        *
 * MODULES: PalmHashDb                                                   */
UInt16 calcHashValue( char *data )
{
	char hash[2] = {0,0};			//these two 8bit chars form the space for our 16bit hash value
	UInt16 hashval;
	int temp,x;

	temp = StrLen(data);
	if( temp % 2 )								//we must check the length so that we know whether the string 
	{
		for( x = 0; x < temp-1; x += 2 )		//ends on an odd or even character
		{
			hash[1] ^= data[x];
			hash[0] ^= data[x+1];
		}
		hash[1] ^= data[x];
	}
	else
	{
		for( x = 0; x < temp; x += 2 )
		{
			hash[1] ^= data[x];
			hash[0] ^= data[x+1];
		}
	}

	hashval = *(UInt16 *)hash;
	hashval = hashval % 32749;
	
	//avoid collision between hash value and sample points (collision is when sample-hash = 0 or sample-hash = prime field)
	//the sample points go as {0, -1, 1, -2, 2... -N/2} for N number of differences so adding half the value of N (or 
	//subtracting it) should avoid collisions
	if( hashval < DIFFERENCES/2 + 1 )
		hashval += DIFFERENCES/2 + 1;
	else if( hashval > 32749 - (DIFFERENCES/2 + 1) )
		hashval -= DIFFERENCES/2 + 1;
	
	return( hashval );
}


/* FUNCTION: calcNewIndex                                                                     *
 * PURPOSE: Given a hash value and a reference to the pre-opened hash database, this function *
 * will do a binary search through the ordered hash values and return the integer index at    *
 * which the given hash value should be inserted to keep the hash list ordered.               *
 * DATE: 9/24/02                                                                              *
 * CONSTANTS: 0 - record index in the hash database that holds the hash value array           *
 * MODULES: PalmHashDb                                                                        */
unsigned short calcNewIndex( UInt16 hashvalue, DmOpenRef hashDb )
{
	UInt16 divisor = 2, numhashes, current, temp;
	MemHandle hashrecord;
	HashRecordType *hashrecorddata;
	
	hashrecord = DmQueryRecord( hashDb, 0 );			//the first record contains all the hashes in the database
	hashrecorddata = MemHandleLock( hashrecord );
	
	numhashes = hashrecorddata[0].docindex;				//the number of hashes in the database is in the dummy first record's document index
	
	if( numhashes == 0 )								//if there are no hash values, the answer is trivial, the new hash value should be the 1st
	{
		MemHandleUnlock( hashrecord );
		return( 1 );
	}
		
	current = (numhashes / divisor) + 1;				//start at the middle of the hash list
	
	while(!((hashrecorddata[current-1].hashval < hashvalue) && (hashrecorddata[current].hashval >= hashvalue) ) )	//if the given hashvalue is not between
	{																												//the current hashvalue and the one 
		divisor *= 2;		//at each loop iteration, jump only half the # of indexes as in the loop before			//before that
		
		temp = current;
		
		if( hashrecorddata[current].hashval < hashvalue )		//if we are too low in the list
		{
			current += (numhashes / divisor);					//jump forward
			if( current == temp )								//when we are close (many iterations into the list) we may find that the integer
			{													//division that jumps us forward actually rounds to 0, in this case, we force a move by 1
				current++;										
				divisor /= 2;									//at this point, the divisor isn't doing any good, so we need to make sure it doesn't wrap to zero
			}
			if( current == numhashes + 1 )						//we must also guard against hitting the end of the list (in which case we are done)
				break;
		}
		else													//if we are too high in the list
		{
			current -= (numhashes / divisor);					//jump backward
			if( current == temp )								//make sure that we moved at least one (see above comment)
			{
				current--;
				divisor /= 2;									//at this point, the divisor isn't doing any good, so we need to make sure it doesn't wrap to zero
			}
		}
	}
	MemHandleUnlock( hashrecord );
	return(current);
}


/* FUNCTION: changeDocPtrAtHashIndex                                                                       *
 * PURPOSE: Given a reference to the pre-opened hash database, an index of a hash record in the hash array *
 * and a document index value, this function will assign the document pointer of the hash record at the    *
 * specified index to equal the given document index value.                                                *
 * DATE: 6/20/02                                                                                           *
 * CONSTANTS: none                                                                                         *
 * MODULES: PalmHashDb                                                                                     */
void changeDocPtrAtHashIndex( DmOpenRef hashDBref, UInt16 index, UInt16 newdocptr )
{
	MemHandle myRecord;
	HashRecordType *hashrecords, modifiedrecord;
	
	myRecord = DmGetRecord( hashDBref, 0 );				//get the record containing the hash array
	hashrecords = MemHandleLock( myRecord );                        

	modifiedrecord.docindex = newdocptr;				//create a hash index to replace the existing one with the new document index pointer
	modifiedrecord.hashval = hashrecords[index].hashval;
	
	DmWrite( hashrecords, index * sizeof(HashRecordType), &modifiedrecord, sizeof(modifiedrecord) );	//update the array with the new index
	
	MemHandleUnlock( myRecord );
	DmReleaseRecord( hashDBref, 0, true );
}


/* FUNCTION: decrementHashIndexesGreaterThan
 * PURPOSE: Given a document index number, this function will go through the hash value array and *
 * decrement each document index that it finds whose value is higher than the given index. This   *
 * function is used to update the hash database pointers to the insertion order database, after a *
 * memo has been deleted. The function returns the hash index whose document pointer matches the  *
 * index passed to the function.                                                                  *
 * DATE: 6/20/02                                                                                  *
 * CONSTANTS: none                                                                                *
 * MODULES PalmHashDb                                                                             *
 * SUGGESTIONS: Try to only call DmWrite once to write the whole array.                           *
 *              Shouldn't x be a larger data type than short?                                     */
UInt16 decrementDocPtrsGreaterThan( DmOpenRef hashDb, UInt16 docnumber )
{
	MemHandle myRecord;
	unsigned short x, index;
	HashRecordType *hashrecords, newrecord;
	
	myRecord = DmGetRecord( hashDb, 0 );								//open the record containing the hash value array
	hashrecords = MemHandleLock( myRecord );							 
	for( x = 1; x < hashrecords[0].docindex+1; x++ )					//for each record in the array
	{
		if( hashrecords[x].docindex > docnumber )							//if its document pointer is greater than the one passed to us
		{
			newrecord.docindex = hashrecords[x].docindex - 1;					//decrement it and write the change to the database
			newrecord.hashval = hashrecords[x].hashval;							
			DmWrite( hashrecords, x * sizeof(HashRecordType), &newrecord, sizeof(HashRecordType) );
		}
		else if( hashrecords[x].docindex == docnumber )						//if its document pointer is the same as the one passed to us
			index = x;															//keep it and return it at the end of the function
	}
	MemHandleUnlock( myRecord );
	DmReleaseRecord( hashDb, 0, true );
	return(x);
}


/* FUNCTION: deleteHashIndex                                                              *
 * PURPOSE: This function will remove a specific index from the hash record array         *
 *          and bump down all the ones after it so that the array will not get fragmented *
 * DATE: 5/28/02                                                                          *
 * CONSTANTS: 0 - record index in the hash database which contains the hash array         *
 * MODULES: PalmHashDb                                                                    */
void deleteHashIndex( UInt16 index, DmOpenRef hashDb )
{
	MemHandle record;
	HashRecordType *hashrecords, newrecord;
	
	record = DmGetRecord( hashDb, 0 );				//start by getting the record that holds the hash array
	hashrecords = MemHandleLock( record );
	
	//take the chunk of memory from after the index to be deleted until the end of the array and copy it
	//back onto the array, only at a starting position of one index back, thus overwriting the index to be deleted
	DmWrite( hashrecords, index * sizeof(HashRecordType), &(hashrecords[index+1]), (hashrecords[0].docindex - index) * sizeof(HashRecordType) );

	newrecord.docindex = -1;		//the "old" last location of the array is now redundant, since the previous
	newrecord.hashval = -1;         //copy moved it back one index, but did not erase the original. Delete it now.
	DmWrite( hashrecords, hashrecords[0].docindex * sizeof(HashRecordType), &newrecord, sizeof(HashRecordType) );
	
	newrecord.docindex = hashrecords[0].docindex - 1;			//update the 0th index (holds the # of hash values in the record) to reflect the change
	newrecord.hashval = 0;
	DmWrite( hashrecords, 0, &newrecord, sizeof(HashRecordType) ); 
	
	MemHandleUnlock( record );						//unlock the record and release it
	DmReleaseRecord( hashDb, 0, true );
}


/* FUNCTION: findIndexWithHashVal                                                             *
 * PURPOSE: Given a hash value and a reference to the pre-opened hash database, this fucntion *
 * performes a binary search of the hash indexes and returns the pointer to the index in the  *
 * document database of the matching hash value (it also returns, through a pointer parameter,*
 * the index in the hash database in which the match was found).                              *
 * DATE: 5/28/02                                                                              *
 * CONSTANTS: 0 - the record in the hash database that holds all the hash records             *
 * MODULES: PalmHashDb                                                                        *
 * SUGGESTION: The assumption here is that any hash value given to this function actually     * 
 * exists in the hash database, there is no error handling, but there should be.              */
UInt16 findIndexWithHashVal( UInt16 value, DmOpenRef hashDb, UInt16 *index )
{
	MemHandle record;
	UInt16 numhashes, currenthash, temp, divisor = 2;
	HashRecordType *hashrecords;
	
	record = DmQueryRecord( hashDb, 0 );			//the first record of the hash database has all the hashes in it
	hashrecords = MemHandleLock( record );
	
	numhashes = hashrecords[0].docindex;			//get the number of hashes from the first "dummy" index
	currenthash = numhashes / divisor;				//start halfway through the list
	
	while( hashrecords[currenthash].hashval != value )		//while we haven't matched the given hash value yet
	{
		divisor *= 2;							//at each loop iteration, jump only half the # of indexes as in the loop before
		temp = currenthash;
		
		if( hashrecords[currenthash].hashval > value )	//if we are too high in the list
		{
			currenthash -= numhashes / divisor;			//jump backward
			if (currenthash == temp )					//when we are close (many iterations into the list) we may find that the integer
				currenthash--;							//division that jumps us forward actually rounds to 0, in this case, we force a move by 1
		}
		else											//if we are too low in the list
		{												
			currenthash += numhashes / divisor;			//jump forward
			if( currenthash == temp )					//make sure that we moved at least one (see above comment)	
				currenthash++;							
		}
	}
	
	MemHandleUnlock( record );
	if( index != NULL )
		*index = currenthash;
	return( hashrecords[currenthash].docindex );
}


/* FUNCTION: findRecordWithDocIndex                                                           *
 * PURPOSE: Given an index value into the document database and a reference to the pre-opened *
 * hash database, this function will find and return the index of the first hash entry in the *
 * database that points to the same index in the document database.                           *
 * DATE: 5/28/02                                                                              *
 * CONSTANTS: 0 - the index of the record in the hash database that has all the hash values   *
 * MODULES: PalmHashDb                                                                        *
 * SUGGESTIONS: How long is an unsigned short? If it is 256, then how can this reference doc  *
 * indexes that are larger than that?                                                         */
unsigned short findRecordWithDocIndex( unsigned short docindex, DmOpenRef hashDb )
{
	UInt16 x;
	MemHandle hashrecord;
	HashRecordType *hashrecorddata;

	hashrecord = DmQueryRecord( hashDb, 0 );
	hashrecorddata = MemHandleLock( hashrecord );

	for( x = 1; x < hashrecorddata[0].docindex + 1; x++ )		//go through linearly
		if( docindex == hashrecorddata[x].docindex )			//stop when you find the match
			break;
			
	MemHandleUnlock( hashrecord );
	return( x );

}


/* FUNCTION: getHashValue                                                                      *
 * PURPOSE: Given a reference to the pre-opened hash database, and the index of the hash entry * 
 * this function will return the value of the hash at that entry.                              *
 * DATE: 6/20/02                                                                               *
 * CONSTANTS: none                                                                             *
 * MODULES: PalmHashDb                                                                         */
UInt16 getHashValue( DmOpenRef hashDb, UInt16 index )
{
	MemHandle myrecord;
	HashRecordType *hashrecord;
	UInt16 hashvalue;
	
	myrecord = DmQueryRecord( hashDb, 0 );			//open the hash array record for reading
	hashrecord = MemHandleLock( myrecord );
	
	hashvalue = hashrecord[index].hashval;			//get the desired hash value
	
	MemHandleUnlock( myrecord );
	
	return( hashvalue );
}


/* FUNCTION: insertHashIndex                                                                             *
 * PURPOSE: Given a complete hashrecord structure, and a reference to the pre-opened hash database, this *
 * function will insert that hashrecord structure into the hash database. The method used here involves  *
 * block memory copying.                                                                                 *
 * DATE: 5/28/02                                                                                         *
 * CONSTANTS: 0 - the index of the record in the hash database that has all the hash values              *
 * MODULES: PalmHashDb                                                                                   */
void insertHashIndex( HashRecordType hashdata, DmOpenRef hashDb )
{
	MemHandle record, newbuffer;
	HashRecordType *hashrecords, *buffer;
	UInt16 index;
	
	//the insert will require an actual temp copy of the array, so allocate some heap for it, and lock it.
	newbuffer = MemHandleNew( NUMMEMOS * sizeof(HashRecordType) );
	buffer = MemHandleLock( newbuffer );
	
	index = calcNewIndex( hashdata.hashval, hashDb );	//find the index in the array to insert the new hash data
	record = DmGetRecord( hashDb, 0 );					//get the hash record array and lock it.
	hashrecords = MemHandleLock( record );

	//first, move all the array before the new index to be inserted to the temporary buffer
	MemMove( buffer, hashrecords, index * sizeof(HashRecordType) );
	
	//second, move all the array after the new index to be inserted into the temp buffer ONE INDEX AHEAD. 
	MemMove( &(buffer[index+1]), &hashrecords[index], (hashrecords[0].docindex - index + 1) * sizeof(HashRecordType));
	
	buffer[index].docindex = hashdata.docindex;				//add the new index to its place in the temp buffer
	buffer[index].hashval = hashdata.hashval;
	buffer[0].docindex = hashrecords[0].docindex + 1;		//update the zeroth element in the temp buffer
	buffer[0].hashval = 0;
	DmWrite( hashrecords, 0, buffer, (buffer[0].docindex + 1) * sizeof(HashRecordType) );	//copy the temp buffer over the original array
																							
	MemHandleUnlock( record );							//unlock and release the hash record
	DmReleaseRecord( hashDb, 0, true );

	MemHandleUnlock( newbuffer );						//free the dynamically allocated memory
	MemHandleFree( newbuffer );
}


/* FUNCTION: openHashDb2                                                 *
 * PURPOSE: Opens the hash values database, if it exists. If it doens't, *
 * then the function creates one of the appropriate size and initializes *
 * it to hold all -1, except the first dummy entry in the array, which   *
 * holds the minimum hash value in its hashvalue field and the number of *
 * hashes in the array in its document index field.                      *
 * DATE: 5/28/02                                                         *
 * CONSTANTS: "SyncHashDB" - name of the hash database                   *
 *            'sTrt' - 32bit creator ID of the database                  *
 *            NUMMEMOS                                                   *
 *            0 - the record in the database that holds the hash array   *
 * MODULES: PalmHashDb                                                   */
DmOpenRef openHashDb2()
{
	LocalID databaseid;
	DmOpenRef hashDb;
	MemHandle hashrecord, newdata;
	HashRecordType *setdata, *temp;
	UInt16 dummy = 0;
	
	databaseid = DmFindDatabase(0, "SyncHashDB");			//find the hash database
	if( !databaseid )										//if it is not found, then create it
	{
		newdata = MemHandleNew( NUMMEMOS * sizeof(HashRecordType) );		//create a memory block of the proper size for the hash database
		setdata = MemHandleLock( newdata );
		MemSet( setdata, NUMMEMOS * sizeof(HashRecordType), -1 );			//set everything in the block to -1
		setdata[0].docindex = 0;											//except the first index (this will hold the minimum hash value 
		setdata[0].hashval = 0;												//and the number of hashes in the array)
		DmCreateDatabase(0, "SyncHashDB", 'sTrt', 0, false );
		databaseid = DmFindDatabase(0, "SyncHashDB");
		hashDb = DmOpenDatabase(0, databaseid, dmModeReadWrite );
		hashrecord = DmNewRecord( hashDb, &dummy, NUMMEMOS * sizeof(HashRecordType) );	//now create the actual record
		temp = MemHandleLock( hashrecord );
		DmWrite( temp, 0, setdata, NUMMEMOS * sizeof( HashRecordType) );				//copy the temporary memory block to the record
		MemHandleUnlock( hashrecord );
		DmReleaseRecord( hashDb, 0, true );
		MemHandleUnlock( newdata );
		MemHandleFree( newdata );
	}
	else
		hashDb = DmOpenDatabase(0, databaseid, dmModeReadWrite );
	
	return( hashDb );
}

#endif