#include "\Palm\PalmMemoDb\PalmMemoDb.h" 


/* FUNCTION: getMemoText                                                                             *
 * PURPOSE: Given a pre-opened references to the document or insertion order document database, and a*
 * memo number, this function will return a pointer to a string that contains the memo text. The last*
 * argument of this function allows callers to specify the location into which the memo text will be *
 * placed; this is useful in situations where global/static variables can't be used.                 *
 * DATE: 6/21/02                                                                                     *
 * CONSTANTS: MEMOSIZE                                                                               *
 * MODULES: PalmMemoDb                                                                               */
char *getMemoText( DmOpenRef docDb, UInt16 memonum, char *localbuffer )
{
	MemHandle myRecord;
	char *record;
	static char buffer[MEMOSIZE];

	myRecord = DmQueryRecord(docDb, memonum);	//since records appear on the list according to their index...
	record = MemHandleLock(myRecord);			//...then the 0th list item is the 0th document index, etc.
	
	if( localbuffer == NULL )				//only use the static storage space, if alternative space hasn't been provided
		StrCopy( buffer, record );			//copy the memo text into a static storage location
	else
		StrCopy( localbuffer, record );
	
	MemHandleUnlock(myRecord);
	
	return( buffer );
}
 

/* FUNCTION: numNonDeletedMemos                                                                     *
 * PURPOSE: Given a pre-openened reference to the document database, this function will return the  *
 * number of non-deleted memos stored in the Palm Memo databases.                                   *
 * DATE: 6/20/02                                                                                    *
 * CONSTANTS: none                                                                                  *
 * MODULES: PalmMemoDb                                                                              */
UInt16 numNonDeletedMemos( DmOpenRef docDb )
{
	return( DmNumRecordsInCategory( docDb, dmAllCategories ) );
}


/* FUNCTION: numTotalMemos                                                                          *
 * PURPOSE: Given a pre-openened reference to the document database, this function will return the  *
 * total number (non-deleted+deleted) of memos stored in the Palm Memo databases.                   *
 * DATE: 6/20/02                                                                                    *
 * CONSTANTS: none                                                                                  *
 * MODULES: PalmMemoDb                                                                              */
UInt16 numTotalMemos( DmOpenRef docDb )
{
	return( DmNumRecords( docDb ));
}


/* FUNCTION: deleteMemoFromDocDb                                                                     *
 * PURPOSE: Given a reference to the pre-opened document database and an index in the document array,*
 * this function will mark that record in the document database as deleted and move that record to   *
 * the end of the database.                                                                          *
 * DATE: 6/20/02                                                                                     *
 * CONSTANTS: none                                                                                   *
 * MODULES: PalmMemoDb                                                                               */
void deleteMemoFromDocDb( DmOpenRef docDb, UInt16 docnumber )
{
		DmDeleteRecord( docDb, docnumber );								
		DmMoveRecord( docDb, docnumber,  numTotalMemos( docDb ) );	
}


/* FUNCTION: findAlphabeticLocation                                                                     *
 * PURPOSE: Given a particular string and a reference to the pre-opened alphabetized document database, *
 * this function will do a binary search and return the index at which the string should be inserted to *
 * maintain alphabetization.                                                                            *
 * DATE: 9/24/02                                                                                        *
 * CONSTANTS: none                                                                                      *
 * MODULES: PalmMemoDb                                                                                  *
 * SUGGESTION: Use the existing numNonDeletedMemos instead of DmNumRecordsInCategory function           */
UInt16 findAlphabeticLocation( char *data, DmOpenRef docDB )
{
	UInt16 divisor = 2, temp, current, numrecs; 
	Int16 compare1, compare2;
	char *docpos, *docposminus1;
	MemHandle record1, record2;
	
	numrecs = DmNumRecordsInCategory(docDB, dmAllCategories);			//get the number of (non-deleted) documents 

	if( numrecs == 0 )			//if there are no documents yet stored, the answer is trivial, the new string should be at index 0
		return( 0 );
	if( numrecs == 1 )			//if there is only one document stored, we do a direct comparison between the new one and the stored one
	{
		record1 = DmQueryRecord( docDB, 0 );
		if( StrCompare( data, MemHandleLock( record1 ) ) <= 0 )		//if the stored one is alphabetically later than the new one
		{
			MemHandleUnlock( record1 );
			return(0);
		}
		else														//if the stored one is alphabetically earlier than the new one
		{
			MemHandleUnlock( record1 );
			return(1);
		}
	}


	current = numrecs / divisor;	//if there is more than just one document in the database, start at the middle of the database

	do								//check to see how the string compares to the document at the current index and at the current-1 index
	{
		record1 = DmQueryRecord( docDB, current );
		record2 = DmQueryRecord( docDB, current - 1);
		docpos = MemHandleLock( record1 );
		docposminus1 = MemHandleLock( record2 );
		compare1 = StrCompare( data, docpos );
		compare2 = StrCompare( data, docposminus1 );
 
		MemHandleUnlock( record1 );
		MemHandleUnlock( record2 );
		
		divisor *= 2;
		
		temp = current;
		if( compare1 > 0 )				//if we are too low in the alphabet
		{
			current += (numrecs / divisor);		//jump ahead
			
			if( current == temp )				//make sure that we actually move up at least one (at high divisor values, numrecs/divisor may round to 0)
			{
				current++;
				divisor /= 2;					//at this point, the divisor isn't doing any good, so we need to make sure it doesn't wrap to zero
			}
			if( current == numrecs )			//if we reach the end of the list, then we are done, and the new string should go at the end
			{
				temp = current;
				break;
			}
		}
		else							//if we are too high in the alphabet
		{
			current -= (numrecs / divisor);		//jump back
			if( current == temp )				//make sure that we actually move down at least one (at high divisor values, numrecs/divisor may round to 0)
			{
				current--;
				divisor /= 2;					//at this point, the divisor isn't doing any good, so we need to make sure it doesn't wrap to zero
			}
			if( current == 0 )					//if we reach the beginning of the list, then we are done
			{									//we just have to see if the string is earlier or later in the alphabet than the first string stored
				if( compare2 <= 0 )
					temp = 0;
				else
					temp = 1;
				return(temp);
			}
		}
	} while( !( (compare1 < 0) && (compare2 >= 0)));	//if we find a condition where the current string is greater and the current-1 string is
														//smaller than the new string, then we are also done and the current location is the right one	
	return(temp);
}


/* FUNCTION: getUnlockedMemoHandle                                                                  * 
 * PURPOSE: Given a reference to the pre-opened document database and an index value, this function *
 * will return the MemHandle that points to the unlocked chunk which contains the memo text at the  *
 * specified index of the document database.                                                        *
 * DATE: 6/20/02                                                                                    *
 * CONSTANTS: none                                                                                  *
 * MODULES: PalmMemoDb                                                                              */
MemHandle getMemoHandle(DmOpenRef docDb, UInt16 docnumber)
{
	MemHandle myRecord;
	char *record;

	myRecord = DmQueryRecord(docDb, docnumber );
	record = MemHandleLock(myRecord);
	MemHandleUnlock(myRecord);
	
	return( myRecord );
}


/* FUNCTION: getMemoTextLength                                                                      *
 * PURPOSE: Given a reference to the pre-opened document database and an index value, this function *
 * will return the length of the memo at the given index just as a strlen function would.           *
 * DATE: 6/20/02                                                                                    *
 * CONSTANTS: none                                                                                  *
 * MODULES: PalmMemoDb                                                                              *
 * SUGGESTIONS: Shouldn't "len" be of a larger type than short?                                     */
UInt16 getMemoTextLength( DmOpenRef docDb, UInt16 index )
{
	MemHandle myRecord;
	char *record;
	short len;
	
	myRecord = DmQueryRecord(docDb, index);
	record = MemHandleLock(myRecord);
	len = StrLen(record);
	MemHandleUnlock(myRecord);

	return( len );
}


/* FUNCTION: removeMemoFromDocDb                                                                      *
 * PURPOSE: Given a reference to the pre-opened document database and a document index, this function *
 * will immediately remove (delete and erase) the record at that index (the data is irretrivable!)    *
 * in the document database.                                                                          *
 * DATE: 6/20/02                                                                                      *
 * CONSTANTS: none                                                                                    *
 * MODULES: PalmMemoDb                                                                                */
void removeMemoFromDocDb( DmOpenRef docDb, UInt16 index )
{
	DmRemoveRecord( docDb, index );
}


/* FUNCTION: removeMemoFromiDocDb                                                                     *
 * PURPOSE: Given a reference to the pre-opened insertion ordered document database and a document    *
 * index, this function will immediately remove (delete and erase) the record at that index (the data *
 * is irretrivable!) in the document database.                                                        *
 * DATE: 6/20/02                                                                                      *
 * CONSTANTS: none                                                                                    *
 * MODULES: PalmMemoDb                                                                                */
void removeMemoFromiDocDb( DmOpenRef idocDb, UInt16 index )
{
	DmRemoveRecord( idocDb, index );
}


/* FUNCTION: saveTestString                                                                                                                 *
 * PURPOSE: Given a memo string, references to the alphabetized document database and the insertion order document database, this function  *
 * will save the memo into the both databases at the appropriate locations. The function will return the index in the insertion order       *
 * database at which the memo string was written.                                                                                           *
 * DATE: 6/11/02                                                                                                                            *
 * CONSTANTS: none                                                                                                                          *
 * MODULES: PalmMemoDb                                                                                                                      */
int saveTestString( char *data, DmOpenRef docDB, DmOpenRef idocDB )
{
	MemHandle docrecord;	
	char *record;
	UInt16 doc, length;	
	
	length = StrLen(data)+1;
	doc = findAlphabeticLocation( data, docDB );			//first of all, find the location for the new memo into alphabetized database
	docrecord = DmNewRecord( docDB, &doc, length );			//create a new record at that location
			
	record = MemHandleLock(docrecord);						//now that we have created the record, write the memo string to it
	DmWrite( record, 0, data, length );
	MemHandleUnlock(docrecord);
	DmReleaseRecord(docDB, doc, true);
	
	doc = DmNumRecords( idocDB );							//next, create a memo location at the end of the insertion ordered documet database
	docrecord = DmNewRecord( idocDB, &doc, length );
	
	record = MemHandleLock(docrecord);						//now that we have created the record, write that memo string to it
	DmWrite( record, 0, data, length );
	MemHandleUnlock(docrecord);
	DmReleaseRecord(idocDB, doc, true);
	
	return( doc );
}


/* FUNCTION: openDocDb                                                             *
 * PURPOSE: Opens the alphabetized document database if it exists. If it doesn't,  *
 * then the function creates one and then opens it.                                *
 * DATE: 5/28/02                                                                   *
 * CONSTANTS: "SyncDocDB" - name of the alphabetized document database             *
 *            'sTrt' - 32bit creator ID of the database                            *
 * MODULES: PalmMemoDb                                                              */
DmOpenRef openDocDb()
{
	LocalID databaseid;
	
	databaseid = DmFindDatabase(0,"SyncDocDB");				//find the document database
	if( !databaseid )										//if it is not found, then create it
	{
		DmCreateDatabase(0, "SyncDocDB", 'sTrt', 0, false );
		databaseid = DmFindDatabase(0,"SyncDocDB");
	}
	return( DmOpenDatabase( 0, databaseid, dmModeReadWrite ));
}

/* FUNCTION: openInsertionDocDb
 * PURPOSE: Opens the insertion order document database, if it exists. If it doesn't *
 * then the function creates one and then opens it.                                  *
 * DATE: 5/28/02                                                                     *
 * CONSTANTS: "SyncIDocDB" - name of the insertion order document database           *
 *            'sTrt' - 32bit creator ID of the database                              *
 * MODULES: PalmMemoDb                                                               */
DmOpenRef openInsertionDocDb()
{
	LocalID databaseid;
	
	databaseid = DmFindDatabase(0,"SyncIDocDB");			//find the document database
	if( !databaseid )										//if it is not found, then create it
	{
		DmCreateDatabase(0, "SyncIDocDB", 'sTrt', 0, false );
		databaseid = DmFindDatabase(0,"SyncIDocDB");
	}
	return( DmOpenDatabase( 0, databaseid, dmModeReadWrite ));
}
