#include "\Common\Constants.h"
#include "\Palm\PalmPolyDb\PalmPolyDb.h"
#include "\Palm\PalmHashDb\PalmHashDb.h"


/* FUNCTION: blockUpdatePolynomialDb                                                                                       *
 * PURPOSE: Given the number of hash records and a list of those hash records, references to the pre-opened polynomial and *
 * alphabetized document databases and an action code, this function will update the polynomial evaluations to reflect the *
 * addition of every hash value in the list to the hash value database. The savings of doing this in a "block" fashion, as *
 * implemented, are in the single locking/writing/unlocking of the polynomial evaluation records.                          *
 * DATE: 6/21/02                                                                                                           *
 * CONSTANTS: DIFFERENCES                                                                                                  *
 *            PROB1                                                                                                        *
 * MODULES: PalmPolyDb                                                                                                     */
void blockUpdatePolynomialDb( UInt16 numhashes, HashRecordType *hashes, DmOpenRef polyDB, DmOpenRef docDB, int code )
{
	MemHandle polyrecord;
	int sample, x, y, hash, dummy;
	UInt16 *polyvals, *recordstart, newpolyvals[DIFFERENCES], size;
	UInt32 newpolyval;
	
	polyrecord = DmGetRecord( polyDB, 0 );		//get the "complete" polynomial list
	recordstart = MemHandleLock(polyrecord);
	polyvals = recordstart + 1;					//the values start after 16bits (which are used to hold the number of documents in the doc. database)
	
	MemMove( newpolyvals, polyvals, DIFFERENCES * sizeof(UInt16) ); //set the newpolyvals samples to the values stored in the database

	for( y = numhashes-1; y >= 0; y-- )			//for each hash value that was added to the hash value database
	{
		hash = hashes[y].hashval;				//get the current hash value
		
		for( x = 0; x < DIFFERENCES; x++ )		//for each of the sample points 
		{
			if( x % 2 )							//find the sample point we are dealing with (0,-1,1,-2,2...)
				sample = -1 * ((x+1)/2);
			else
				sample = x/2;
				
			newpolyval = newpolyvals[x];		//start with copying the existing evaluation
			
			dummy = (sample) - hash;			//calculate the new multiplyer
			if( dummy < 0 )
				dummy += 32749;
				
			newpolyval *= dummy;				//multiply [add the entry] to the list

			newpolyval = newpolyval % 32749;	//adjust the new sample to fit the field

			newpolyvals[x] = newpolyval;		//write the sample to an array that holds all the new sample values
		}
	}
	
	DmWrite( recordstart, sizeof(UInt16), &newpolyvals, sizeof(newpolyvals) );	//write the new array on top of the old one
	dummy = numTotalMemos( docDB );							//also, add as the first 16bits, the number of documents in the document database
	DmWrite( recordstart, 0, &dummy, sizeof(UInt16) );
	MemHandleUnlock(polyrecord);
	DmReleaseRecord(polyDB, 0, true);

	polyrecord = DmGetRecord( polyDB, 1 );			//write a small portion of the new array to the first "fragmented" record
	recordstart = MemHandleLock( polyrecord );		
	DmWrite( recordstart, sizeof(UInt16), &newpolyvals, PROB1 * sizeof( UInt16 ) );
	DmWrite( recordstart, 0, &dummy, sizeof(UInt16) );		//the first fragmented record (for ProbSync) also needs to hold
	MemHandleUnlock( polyrecord );							//the number of documents in the documents database.
	DmReleaseRecord( polyDB, 1, true );		
	size = PROB1*2;
	
	for( x = 2; size <= DIFFERENCES/2; x++ )				//for each "fragmented" record after the first, store part of the new array
	{
		polyrecord = DmGetRecord( polyDB, x );
		polyvals = MemHandleLock( polyrecord );
		DmWrite( polyvals, 0, &size, sizeof(UInt16) );		//in the first 16bits, store the number of records this fragment will hold
		DmWrite( polyvals, sizeof(UInt16), &(newpolyvals[size-PROB1]), size * sizeof( UInt16 ) );
		MemHandleUnlock( polyrecord );
		DmReleaseRecord( polyDB, x, true );
		size *= 2;
	}
}


/* FUNCTION: EuclideanInverse                                                      *
 * PURPOSE: Given a number and the size of the field, this function will calculate *
 * that number's Euclidean Inverse within that field.                              *
 * DATE: 5/28/02                                                                   *
 * CONSTANTS: none                                                                 *
 * MODULES: PalmPolyDb                                                             */
int EuclideanInverse( int number, int field )
{
	int a_0, a_1, a_2;				//this algorithm is just a programmatized version of the algorithm presented at
	int b_0=1, b_1, b_2;			//http://www-math.cudenver.edu/~wcherowi/courses/m5410/exeucalg.html
	int x_0, x_1, x_2;
	int y_0, y_1, y_2;
	long p_0, p_1, p_2;
	long temp = -1;
	
	number = number % field;
	
	y_2 = field;
	x_2 = number;
	a_2 = y_2 / x_2;
	b_2 = y_2 % (x_2 * a_2);
	p_2 = 0;
	
	y_1 = x_2;
	x_1 = b_2;
	a_1 = y_1 / x_1;
	b_1 = y_1 % (x_1 * a_1);
	p_1 = 1;
	
	y_0 = x_1;
	x_0 = b_1;
	if( x_0 != 0 )			//must check for algorithm runs that are already completed
		while( b_0 != 0 )
		{
			a_0 = y_0 / x_0;
			b_0 = y_0 % (x_0 * a_0);
			temp = (p_2 - p_1 * a_2);	
			while( temp < 0 )
				temp += field;
			p_0 = temp % field;
		
			p_2 = p_1;
			p_1 = p_0;
		
			a_2 = a_1;
			a_1 = a_0;
		
			b_2 = b_1;
			b_1 = b_0;
		
			x_2 = x_1;
			x_1 = x_0;
			x_0 = b_1;
		
			y_2 = y_1;
			y_1 = y_0;
			y_0 = x_1;
		}
	
	temp = p_2 - p_1 * a_2;
	while( temp < 0 )
		temp += field;
	return( temp % field );
}


/* FUNCTION: openPolyDb
 * PURPOSE: This function opens the characteristic polynomial database if it exists. *
 * If it doesn't, then the function creates the database, as well as the records in  *
 * the database (there are two types: one "complete" for CPISync which holds all the *
 * polynomials; and a number of "fragmented" ones, that hold only parts of the       *
 * complete list of polynomials.                                                     *
 * DATE: 5/28/02                                                                     *
 * CONSTANTS:  DIFFERENCES                                                           *
 *             SyncPolyDB - the name of the polynomial database                      *
 *            'sTrt' - 32bit creator ID of the database                              *
 *             PROB1                                                                 *
 * MODULES: PalmPolyDb                                                               */
DmOpenRef openPolyDb()
{
	LocalID databaseid;
	unsigned short x;
	MemHandle polyrecord;
	UInt16 dummy = 1, numrecs = 1, multiplier = 1;
	DmOpenRef poly;
	UInt16 newpolyvals[DIFFERENCES], size = 0;
	void *rec;
	
	databaseid = DmFindDatabase(0,"SyncPolyDB");
	if( !databaseid )										//find the characterisic polynomial db (if it exists)
	{														//if it doesn't... create it and init. first #DIFFERENCE to 1
		DmCreateDatabase(0, "SyncPolyDB", 'sTrt', 0, false );
		databaseid = DmFindDatabase(0, "SyncPolyDB");
		numrecs = 0;
	}
	poly = DmOpenDatabase(0,databaseid, dmModeReadWrite );
	
	if( numrecs == 0 )										//if we have just created the polynomial database, then we need to initialize its entries
	{
		for( x = 0; x < DIFFERENCES; x++ )			//set the complete record to all ones.
			newpolyvals[x] = 1;
		
		size = DIFFERENCES;
		for( x = 0; size <= DIFFERENCES; x++ )		//create a bunch of fragmented records each holding PROB1 times more records than the one before
		{
			polyrecord = DmNewRecord( poly, &x, ( size + 1 ) * sizeof(UInt16) );
			rec = MemHandleLock(polyrecord);
			DmWrite( rec, sizeof(UInt16), &newpolyvals, size * sizeof(UInt16) );
			dummy = 0;
			if( (x == 0) || (x == 1) )						//the first two records (the "complete" one and the first fragmented one) 
				DmWrite( rec, 0, &dummy, sizeof(UInt16) );  //hold the number of documents in the document database in 0'th index
			else											//all other fragmented records hold their own size in 0th index
				DmWrite( rec, 0, &size, sizeof(UInt16) );
			MemHandleUnlock( polyrecord );
			DmReleaseRecord( poly, x, false );
						
			size = PROB1 * multiplier;
			multiplier *= 2;
		}
	}
	return(poly);
}


/* FUNCTION: updateDocumentCount
 * PURPOSE: The first 16bits of the first two records of the polynomial database *
 * are reserved for holding the count of the number of documents on the Palm.    *
 * Given a reference to the pre-opened document and polynomial databases, this   *
 * function will put that count into both the reserved locations.                *
 * DATE: 5/23/02                                                                 *
 * CONSTANTS: 0 - the location of the record in the polynomial database that     *
 *                contains all the evaluations.                                  *
 *            1 - the location of the first record in the polynomial database    *
 *                that holds the fragmented list of polynomial evaluations.      *
 * MODULES: PalmPolyDb                                                           *
 * SUGGESTIONS: Shouldn't the document count reflect the total number of         *
 * documents in the document database, rather than just the non-deleted ones?    *
 *              Rename this function to "updateDocumentCountinPolyDatabase"      */
void updateDocumentCount( DmOpenRef docDB, DmOpenRef polyDB )
{
	MemHandle myRecord;											//since the assupmtion is that all currently
	UInt16 *poly, dummy;                                        //deleted records will be purged, the document
	                                                            //count must be modified to reflect that.
	myRecord = DmGetRecord( polyDB, 0 );						
	poly = MemHandleLock(myRecord);								
	dummy = DmNumRecordsInCategory( docDB, dmAllCategories );	//put the number of records in a temp location
	DmWrite( poly, 0, &dummy, sizeof(UInt16) );					//write that value to the database
	MemHandleUnlock(myRecord);
	DmReleaseRecord( polyDB, 0, true );
		
	myRecord = DmGetRecord( polyDB, 1 );
	poly = MemHandleLock(myRecord);
	DmWrite( poly, 0, &dummy, sizeof(UInt16) );
	MemHandleUnlock(myRecord);
	DmReleaseRecord( polyDB, 1, true );
}


/* FUNCTION: updatePolynomialDb                                                                                                  *
 * PURPOSE: Given a hash value and a reference to the polynomial database, and a reference to the alphabetized document database *
 * and an action code, this function will update the polynomial evaluations to reflect the (action code = 0) addition or         *
 * (action code = 1) deletion of the hash value to/from the hash value database. This function handles both the "complete" and   *
 * "fragmented" polynomial evaluation records.                                                                                   *
 * DATE: 6/20/02                                                                                                                 *
 * CONSTANTS: DIFFERENCES                                                                                                        *
 *            32749 - Prime field                                                                                                *
 *            PROB1                                                                                                              *
 *            2 - The multiplier of size between two fragmented records                                                          *
 * MODULES: PalmPolyDb                                                                                                           */
void updatePolynomialDb( int hash, DmOpenRef polyDB, DmOpenRef docDB, int code )
{
	MemHandle polyrecord;
	int sample, x, dummy;
	UInt16 *polyvals, *recordstart, newpolyvals[DIFFERENCES], size;
	long newpolyval;
	
	polyrecord = DmGetRecord( polyDB, 0 );		//get the "complete" polynomial list
	recordstart = MemHandleLock(polyrecord);
	polyvals = recordstart + 1;					//the values start after 16bits (which are used to hold the number of documents in the doc. database)
	
	for( x = 0; x < DIFFERENCES; x++ )			//create as many evaluations as are expected by the upper bound DIFFERENCES
	{
		if( x % 2 )								//find the sample point we are dealing with (0,-1,1,-2,2...)
			sample = -1 * ((x+1)/2);
		else
			sample = x/2;
			
		newpolyval = polyvals[x];				//start with copying the existing evaluation
		
		dummy = (sample) - hash;				//calculate the new multiplyer
		if( dummy < 0 )
			dummy += 32749;

		if( !code  )							//if code = 0, then multiply [add the entry] to the list
			newpolyval *= dummy;				
		else
			newpolyval *= EuclideanInverse(dummy,32749);	//otherwise, multiply by inverse [remove the entry] from the list

		if( newpolyval < 0 )					//adjust the new sample to fit the field
			newpolyval += 32749;
		else
			newpolyval = newpolyval % 32749;
			
		newpolyvals[x] = newpolyval;			//write the sample to an array that holds all the new sample values
	}
		
	DmWrite( recordstart, sizeof(UInt16), &newpolyvals, sizeof(newpolyvals) );	//write the new array on top of the old one
	dummy = numTotalMemos( docDB );							//also, add as the first 16bits, the number of documents in the document database
	DmWrite( recordstart, 0, &dummy, sizeof(UInt16) );
	MemHandleUnlock(polyrecord);
	DmReleaseRecord(polyDB, 0, true);

	polyrecord = DmGetRecord( polyDB, 1 );			
	recordstart = MemHandleLock( polyrecord );				//write a small portion of the new array to the first "fragmented" record
	DmWrite( recordstart, sizeof(UInt16), &newpolyvals, PROB1 * sizeof( UInt16 ) );		
	DmWrite( recordstart, 0, &dummy, sizeof(UInt16) );		//the first fragmented record (for ProbSync) also needs to hold
	MemHandleUnlock( polyrecord );							//the number of documents in the documents database.
	DmReleaseRecord( polyDB, 1, true );
	size = PROB1*2;
	
	for( x = 2; size <= DIFFERENCES/2; x++ )				//for each "fragmented" record after the first, store part of the new array
	{														
		polyrecord = DmGetRecord( polyDB, x );
		polyvals = MemHandleLock( polyrecord );
		DmWrite( polyvals, 0, &size, sizeof(UInt16) );		//in the first 16bits, store the number of records this fragment will hold
		DmWrite( polyvals, sizeof(UInt16), &(newpolyvals[size-PROB1]), size * sizeof( UInt16 ) );
		MemHandleUnlock( polyrecord );
		DmReleaseRecord( polyDB, x, true );
		size *= 2;
	}
}


